@@ -2,6 +2,7 @@
 #include <pygobject.h>
 #include "pygnomevfs-private.h"
 #include <libgnomevfs/gnome-vfs-utils.h>
+#include <libgnomevfs/gnome-vfs-find-directory.h>
 #include <pyorbit.h>
 
 static GHashTable *monitor_hash;
@@ -321,13 +322,17 @@ pygvfs_create(PyObject *self, PyObject *args, PyObject *kwargs)
 				     &perm))
 	return NULL;
 
-    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type))
+    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_create_uri(&handle, pygnome_vfs_uri_get(uri),
 				      open_mode, exclusive, perm);
-    else if (PyString_Check(uri))
+        pyg_end_allow_threads;
+    } else if (PyString_Check(uri)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_create(&handle, PyString_AsString(uri),
 				  open_mode, exclusive, perm);
-    else {
+        pyg_end_allow_threads;
+    } else {
 	PyErr_SetString(PyExc_TypeError,
 			"uri must be a gnomevfs.URI or a string");
 	return NULL;
@@ -354,13 +359,17 @@ pygvfs_get_file_info(PyObject *self, PyObject *args, PyObject *kwargs)
 	return NULL;
 
     finfo = gnome_vfs_file_info_new();
-    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type))
+    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_get_file_info_uri(pygnome_vfs_uri_get(uri), finfo,
 					     options);
-    else if (PyString_Check(uri))
+        pyg_end_allow_threads;
+    } else if (PyString_Check(uri)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_get_file_info(PyString_AsString(uri), finfo,
 					 options);
-    else {
+        pyg_end_allow_threads;
+    } else {
 	PyErr_SetString(PyExc_TypeError,
 			"uri must be a gnomevfs.URI or a string");
 	gnome_vfs_file_info_unref(finfo);
@@ -390,13 +399,17 @@ pygvfs_set_file_info(PyObject *self, PyObject *args, PyObject *kwargs)
                                      &mask))
 	return NULL;
 
-    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type))
+    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_set_file_info_uri(pygnome_vfs_uri_get(uri),
                                              finfo->finfo, mask);
-    else if (PyString_Check(uri))
+        pyg_end_allow_threads;
+    } else if (PyString_Check(uri)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_set_file_info(PyString_AsString(uri),
                                          finfo->finfo, mask);
-    else {
+        pyg_end_allow_threads;
+    } else {
 	PyErr_SetString(PyExc_TypeError,
 			"uri must be a gnomevfs.URI or a string");
 	return NULL;
@@ -454,12 +467,16 @@ pygvfs_make_directory(PyObject *self, PyObject *args, PyObject *kwargs)
 				     &uri, &perm))
 	return NULL;
 
-    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type))
+    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_make_directory_for_uri(pygnome_vfs_uri_get(uri),
 						  perm);
-    else if (PyString_Check(uri))
+        pyg_end_allow_threads;
+    } else if (PyString_Check(uri)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_make_directory(PyString_AsString(uri), perm);
-    else {
+        pyg_end_allow_threads;
+    } else {
 	PyErr_SetString(PyExc_TypeError,
 			"uri must be a gnomevfs.URI or a string");
 	return NULL;
@@ -484,11 +501,15 @@ pygvfs_remove_directory(PyObject *self, PyObject *args, PyObject *kwargs)
 				     &uri))
 	return NULL;
 
-    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type))
+    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_remove_directory_from_uri(pygnome_vfs_uri_get(uri));
-    else if (PyString_Check(uri))
+        pyg_end_allow_threads;
+    } else if (PyString_Check(uri)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_remove_directory(PyString_AsString(uri));
-    else {
+        pyg_end_allow_threads;
+    } else {
 	PyErr_SetString(PyExc_TypeError,
 			"uri must be a gnomevfs.URI or a string");
 	return NULL;
@@ -512,11 +533,15 @@ pygvfs_unlink(PyObject *self, PyObject *args, PyObject *kwargs)
 				     "O:gnomevfs.unlink", kwlist, &uri))
 	return NULL;
 
-    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type))
+    if (PyObject_TypeCheck(uri, &PyGnomeVFSURI_Type)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_unlink_from_uri(pygnome_vfs_uri_get(uri));
-    else if (PyString_Check(uri))
+        pyg_end_allow_threads;
+    } else if (PyString_Check(uri)) {
+        pyg_begin_allow_threads;
 	result = gnome_vfs_unlink(PyString_AsString(uri));
-    else {
+        pyg_end_allow_threads;
+    } else {
 	PyErr_SetString(PyExc_TypeError,
 			"uri must be a gnomevfs.URI or a string");
 	return NULL;
@@ -541,11 +566,15 @@ pygvfs_exists(PyObject *self, PyObject *args, PyObject *kwargs)
 				     kwlist, &py_uri))
 	return NULL;
 
-    if (PyObject_TypeCheck(py_uri, &PyGnomeVFSURI_Type))
+    if (PyObject_TypeCheck(py_uri, &PyGnomeVFSURI_Type)) {
+        pyg_begin_allow_threads;
 	uri = gnome_vfs_uri_ref(pygnome_vfs_uri_get(py_uri));
-    else if (PyString_Check(py_uri))
+        pyg_end_allow_threads;
+    } else if (PyString_Check(py_uri)) {
+        pyg_begin_allow_threads;
 	uri = gnome_vfs_uri_new(PyString_AsString(py_uri));
-
+        pyg_end_allow_threads;
+    }
     if (!uri) {
 	PyErr_SetString(PyExc_TypeError,
 			"uri must be a gnomevfs.URI or a string");
@@ -566,7 +595,9 @@ pygvfs_get_mime_type(PyObject *self, PyObject *args)
 			 &text_uri))
 	return NULL;
 
+    pyg_begin_allow_threads;
     mime = gnome_vfs_get_mime_type(text_uri);
+    pyg_end_allow_threads;
     if (mime)
 	return PyString_FromString(mime);
     else {
@@ -588,7 +619,9 @@ pygvfs_get_mime_type_for_data(PyObject *self, PyObject *args)
 	return NULL;
     if (data_size1 != -G_MAXINT)
 	PyErr_Warn(PyExc_DeprecationWarning, "ignoring deprecated argument data_size");
+    pyg_begin_allow_threads;
     mime = gnome_vfs_get_mime_type_for_data(data, data_size);
+    pyg_end_allow_threads;
     if (mime)
 	return PyString_FromString(mime);
     else {
@@ -761,9 +794,11 @@ pygvfs_monitor_add(PyObject *self, PyObject *args)
     cunote->data = extra;
     Py_INCREF(cunote->func);
     Py_XINCREF(cunote->data);
+    pyg_begin_allow_threads;
     result = gnome_vfs_monitor_add(&handle, text_uri, monitor_type,
 		   (GnomeVFSMonitorCallback)pygvfs_monitor_marshal,
 		   cunote);
+    pyg_end_allow_threads;
 
     if (pygnome_vfs_result_check(result)) {
 	return NULL;
@@ -815,7 +850,11 @@ pygvfs_read_entire_file(PyObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "s:gnomevfs.read_entire_file", &uri))
 	return NULL;
+
+    pyg_begin_allow_threads;
     result = gnome_vfs_read_entire_file (uri, &file_size, &file_contents);
+    pyg_end_allow_threads;
+
     if (pygnome_vfs_result_check(result))
 	return NULL;
     rv = PyString_FromStringAndSize(file_contents, file_size);
@@ -862,13 +901,15 @@ pygvfs_mime_get_default_application(PyObject *self, PyObject *args)
 }
 
 
-static gint
+gint
 pygvfs_xfer_progress_callback(GnomeVFSXferProgressInfo *info, gpointer _data)
 {
     PyGVFSCustomNotify *data = _data;
     PyObject *py_info, *callback_return;
     gint retval;
+    PyGILState_STATE state;
 
+    state = pyg_gil_state_ensure();
     py_info = pygnome_vfs_xfer_progress_info_new(info);
     if (data->data)
 	callback_return = PyObject_CallFunction(data->func, "OO", py_info, data->data);
@@ -881,14 +922,22 @@ pygvfs_xfer_progress_callback(GnomeVFSXferProgressInfo *info, gpointer _data)
        * python wrapper to the C structure */
     pygnome_vfs_xfer_progress_info_set(py_info, NULL);
 
-    if (callback_return == NULL)
+    Py_DECREF(py_info);
+
+    if (callback_return == NULL) {
+        PyErr_Print();
+        pyg_gil_state_release(state);
 	return GNOME_VFS_XFER_ERROR_ACTION_ABORT;
+    }
     if (!PyInt_Check(callback_return)) {
 	PyErr_SetString(PyExc_TypeError, "progress callback must return an int");
+        PyErr_Print();
+        pyg_gil_state_release(state);
 	return GNOME_VFS_XFER_ERROR_ACTION_ABORT;
     }
     retval = PyInt_AsLong(callback_return);
     Py_DECREF(callback_return);
+    pyg_gil_state_release(state);
     return retval;
 }
 
@@ -903,7 +952,7 @@ pygvfs_xfer_uri(PyObject *self, PyObject *args, PyObject *kwargs)
     PyGVFSCustomNotify custom_data = {NULL, NULL};
     GnomeVFSResult result;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!O!iiiO|O:gnomevfs.xfer_uri",
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!O!iii|OO:gnomevfs.xfer_uri",
 				     kwlist,
 				     &PyGnomeVFSURI_Type, &source_uri,
 				     &PyGnomeVFSURI_Type, &target_uri,
@@ -911,15 +960,26 @@ pygvfs_xfer_uri(PyObject *self, PyObject *args, PyObject *kwargs)
 				     &custom_data.func, &custom_data.data))
 	return NULL;
 
-    if (!PyCallable_Check(custom_data.func)) {
+    if (custom_data.func == Py_None)
+        custom_data.func = NULL;
+    if (!custom_data.func) {
+        if (error_mode == GNOME_VFS_XFER_ERROR_MODE_QUERY) {
+            PyErr_SetString(PyExc_ValueError, "callback is required with QUERY error mode");
+            return NULL;
+        }
+    } else if (!PyCallable_Check(custom_data.func)) {
 	PyErr_SetString(PyExc_TypeError, "progress_callback must be callable");
 	return NULL;
     }
 
+    pyg_begin_allow_threads;
     result = gnome_vfs_xfer_uri(pygnome_vfs_uri_get(source_uri),
 				pygnome_vfs_uri_get(target_uri),
 				xfer_options,  error_mode, overwrite_mode,
-				pygvfs_xfer_progress_callback, &custom_data);
+				(custom_data.func?
+                                 pygvfs_xfer_progress_callback : NULL),
+                                &custom_data);
+    pyg_end_allow_threads;
     if (pygnome_vfs_result_check(result))
 	return NULL;
 
@@ -927,7 +987,7 @@ pygvfs_xfer_uri(PyObject *self, PyObject *args, PyObject *kwargs)
     return Py_None;
 }
 
-static gboolean
+gboolean
 _pygvfs_uri_sequence_to_glist(PyObject *seq, GList **list)
 {
     int len, i;
@@ -963,7 +1023,7 @@ pygvfs_xfer_uri_list(PyObject *self, PyObject *args, PyObject *kwargs)
     PyGVFSCustomNotify custom_data = {NULL, NULL};
     GnomeVFSResult result;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOiiiO|O:gnomevfs.xfer_uri_list",
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOiii|OO:gnomevfs.xfer_uri_list",
 				     kwlist,
 				     &py_source_uri_list, &py_target_uri_list,
 				     &xfer_options, &error_mode, &overwrite_mode,
@@ -982,17 +1042,28 @@ pygvfs_xfer_uri_list(PyObject *self, PyObject *args, PyObject *kwargs)
 	return NULL;
     }
 
-
-    if (!PyCallable_Check(custom_data.func)) {
+    if (custom_data.func == Py_None)
+        custom_data.func = NULL;
+    if (!custom_data.func) {
+        if (error_mode == GNOME_VFS_XFER_ERROR_MODE_QUERY) {
+            PyErr_SetString(PyExc_ValueError, "callback is required with QUERY error mode");
+            return NULL;
+        }
+    } else if (!PyCallable_Check(custom_data.func)) {
 	PyErr_SetString(PyExc_TypeError, "progress_callback must be callable");
 	g_list_free(source_uri_list);
 	g_list_free(target_uri_list);
 	return NULL;
     }
 
+    pyg_begin_allow_threads;
     result = gnome_vfs_xfer_uri_list(source_uri_list, target_uri_list,
 				     xfer_options,  error_mode, overwrite_mode,
-				     pygvfs_xfer_progress_callback, &custom_data);
+				     (custom_data.func?
+                                      pygvfs_xfer_progress_callback : NULL),
+                                     &custom_data);
+    pyg_end_allow_threads;
+
     g_list_free(source_uri_list);
     g_list_free(target_uri_list);
     if (pygnome_vfs_result_check(result))
@@ -1014,7 +1085,7 @@ pygvfs_xfer_delete_list(PyObject *self, PyObject *args, PyObject *kwargs)
     PyGVFSCustomNotify custom_data = {NULL, NULL};
     GnomeVFSResult result;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OiiO|O:gnomevfs.xfer_delete_list",
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "Oii|OO:gnomevfs.xfer_delete_list",
 				     kwlist,
 				     &py_delete_uri_list,
 				     &error_mode, &xfer_options,
@@ -1027,16 +1098,27 @@ pygvfs_xfer_delete_list(PyObject *self, PyObject *args, PyObject *kwargs)
 	return NULL;
     }
 
-
-    if (!PyCallable_Check(custom_data.func)) {
+    if (custom_data.func == Py_None)
+        custom_data.func = NULL;
+    if (!custom_data.func) {
+        if (error_mode == GNOME_VFS_XFER_ERROR_MODE_QUERY) {
+            PyErr_SetString(PyExc_ValueError, "callback is required with QUERY error mode");
+            return NULL;
+        }
+    } else if (!PyCallable_Check(custom_data.func)) {
 	PyErr_SetString(PyExc_TypeError, "progress_callback must be callable");
 	g_list_free(delete_uri_list);
 	return NULL;
     }
 
+    pyg_begin_allow_threads;
     result = gnome_vfs_xfer_delete_list(delete_uri_list,
 					error_mode, xfer_options,
-					pygvfs_xfer_progress_callback, &custom_data);
+					(custom_data.func?
+                                         pygvfs_xfer_progress_callback : NULL),
+                                        &custom_data);
+    pyg_end_allow_threads;
+
     g_list_free(delete_uri_list);
     if (pygnome_vfs_result_check(result))
 	return NULL;
@@ -1574,6 +1656,242 @@ pygvfs_connect_to_server(PyObject *self, PyObject *args, PyObject *kwargs)
     return Py_None;
 }
 
+#define pygvfs_generic_str_map_func(name, argname)              \
+static PyObject *                                               \
+pygvfs_##name(PyObject *self, PyObject *args, PyObject *kwargs) \
+{                                                               \
+    static char *kwlist[] = { #argname, NULL };                 \
+    char *str;                                                  \
+    PyObject *retval;                                           \
+                                                                \
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,              \
+				     "s:gnomevfs."#name,        \
+				     kwlist, &str))             \
+	return NULL;                                            \
+    str = gnome_vfs_##name(str);                                \
+    if (!str) {                                                 \
+        PyErr_SetString(PyExc_RuntimeError, "unknown error");   \
+        return NULL;                                            \
+    }                                                           \
+    retval = PyString_FromString(str);                          \
+    g_free(str);                                                \
+    return retval;                                              \
+}
+
+pygvfs_generic_str_map_func(escape_string, string);
+pygvfs_generic_str_map_func(escape_path_string, path);
+pygvfs_generic_str_map_func(escape_host_and_path_string, string);
+pygvfs_generic_str_map_func(escape_slashes, string);
+
+static PyObject *
+pygvfs_escape_set(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "string", "match_set", NULL };
+    char *str, *str1;
+    PyObject *retval;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "ss:gnomevfs.escape_set",
+				     kwlist, &str, &str1))
+	return NULL;
+    str = gnome_vfs_escape_set(str, str1);
+    if (!str) {
+        PyErr_SetString(PyExc_RuntimeError, "unknown error");
+        return NULL;
+    }
+    retval = PyString_FromString(str);
+    g_free(str);
+    return retval;
+}
+
+static PyObject *
+pygvfs_unescape_string(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "escaped_string", "illegal_characters", NULL };
+    char *str, *str1;
+    PyObject *retval;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "ss:gnomevfs.unescape_string",
+				     kwlist, &str, &str1))
+	return NULL;
+    str = gnome_vfs_escape_set(str, str1);
+    if (!str) {
+        PyErr_SetString(PyExc_RuntimeError, "unknown error");
+        return NULL;
+    }
+    retval = PyString_FromString(str);
+    g_free(str);
+    return retval;
+}
+
+pygvfs_generic_str_map_func(make_uri_canonical, uri);
+pygvfs_generic_str_map_func(make_path_name_canonical, path);
+pygvfs_generic_str_map_func(unescape_string_for_display, escaped);
+pygvfs_generic_str_map_func(get_local_path_from_uri, uri);
+pygvfs_generic_str_map_func(get_uri_from_local_path, local_full_path);
+
+static PyObject *
+pygvfs_is_executable_command_string(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "command_string", NULL };
+    char *str;
+    gboolean retval;
+    PyObject *py_retval;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "s:gnomevfs.is_executable_command_string",
+				     kwlist, &str))
+	return NULL;
+    retval = gnome_vfs_is_executable_command_string(str);
+    py_retval = retval? Py_True : Py_False;
+    Py_INCREF(py_retval);
+    return py_retval;
+}
+
+static PyObject *
+pygvfs_get_volume_free_space(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "vfs_uri", NULL };
+    PyObject *py_uri;
+    GnomeVFSFileSize size = 0;
+    GnomeVFSResult result;
+    PyObject *py_retval;
+    
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "O!:gnomevfs.get_volume_free_space",
+				     kwlist,
+                                     &PyGnomeVFSURI_Type, &py_uri))
+	return NULL;
+    result = gnome_vfs_get_volume_free_space(pygnome_vfs_uri_get(py_uri), &size);
+    if (pygnome_vfs_result_check(result))
+        return NULL;
+    py_retval = PyLong_FromUnsignedLongLong(size);
+    return py_retval;
+}
+
+pygvfs_generic_str_map_func(icon_path_from_filename, filename);
+
+static PyObject *
+pygvfs_open_fd(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "filedes", NULL };
+    int filedes;
+    GnomeVFSResult result;
+    PyObject *py_retval;
+    GnomeVFSHandle *handle = NULL;
+    
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "i:gnomevfs.open_fd",
+				     kwlist, &filedes))
+	return NULL;
+    result = gnome_vfs_open_fd(&handle, filedes);
+    if (pygnome_vfs_result_check(result))
+        return NULL;
+    py_retval = pygnome_vfs_handle_new(handle);
+    return py_retval;
+}
+
+static PyObject *
+pygvfs_is_primary_thread(void)
+{
+    gboolean retval;
+    PyObject *py_retval;
+
+    retval = gnome_vfs_is_primary_thread();
+    py_retval = retval? Py_True : Py_False;
+    Py_INCREF(py_retval);
+    return py_retval;
+}
+
+pygvfs_generic_str_map_func(format_uri_for_display, uri);
+pygvfs_generic_str_map_func(make_uri_from_input, uri);
+
+static PyObject *
+pygvfs_make_uri_from_input_with_dirs(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "escaped_string", "illegal_characters", NULL };
+    char *str;
+    int dirs;
+    PyObject *retval;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "si:gnomevfs.make_uri_from_input_with_dirs",
+				     kwlist, &str, &dirs))
+	return NULL;
+    str = gnome_vfs_make_uri_from_input_with_dirs(str, dirs);
+    if (!str) {
+        PyErr_SetString(PyExc_RuntimeError, "unknown error");
+        return NULL;
+    }
+    retval = PyString_FromString(str);
+    g_free(str);
+    return retval;
+}
+
+pygvfs_generic_str_map_func(make_uri_canonical_strip_fragment, uri);
+
+static PyObject *
+pygvfs_uris_match(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "uri_1", "uri_2", NULL };
+    char *uri1, *uri2;
+    gboolean retval;
+    PyObject *py_retval;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "ss:gnomevfs.uris_match",
+				     kwlist, &uri1, uri2))
+	return NULL;
+    retval = gnome_vfs_uris_match(uri1, uri2);
+    py_retval = retval? Py_True : Py_False;
+    Py_INCREF(py_retval);
+    return py_retval;
+}
+
+pygvfs_generic_str_map_func(get_uri_scheme, uri);
+pygvfs_generic_str_map_func(make_uri_from_shell_arg, uri);
+
+static PyObject *
+pygvfs_url_show(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "url", "env", NULL };
+    char *url, **env;
+    GnomeVFSResult result;
+    PyObject *py_env = NULL;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "s|O!:gnomevfs.url_show",
+				     kwlist, &url, &PyList_Type, &py_env))
+	return NULL;
+
+    if (!py_env)
+        env = NULL;
+    else {
+        int len = PyList_Size(py_env);
+        int i;
+        env = g_new(char *, len + 1);
+        for (i = 0; i < len; ++i) {
+            PyObject *item = PyList_GET_ITEM(py_env, i);
+            if (!PyString_Check(item)) {
+                PyErr_SetString(PyExc_TypeError, "second argument (env) "
+                                "must be a list of strings");
+                g_free(env);
+                return NULL;
+            }
+            env[i] = PyString_AsString(item);
+        }
+        env[len] = NULL;
+    }
+
+    result = gnome_vfs_url_show_with_env(url, env);
+    if (env) g_free(env);
+    if (pygnome_vfs_result_check(result))
+        return NULL;
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
 
 static PyMethodDef pygnomevfs_functions[] = {
     { "create", (PyCFunction)pygvfs_create, METH_VARARGS|METH_KEYWORDS },
@@ -1654,6 +1972,55 @@ static PyMethodDef pygnomevfs_functions[] = {
       METH_VARARGS|METH_KEYWORDS },
     { "connect_to_server", (PyCFunction)pygvfs_connect_to_server,
       METH_VARARGS|METH_KEYWORDS },
+    { "escape_string", (PyCFunction)pygvfs_escape_string,
+      METH_VARARGS|METH_KEYWORDS },
+    { "escape_path_string", (PyCFunction)pygvfs_escape_path_string,
+      METH_VARARGS|METH_KEYWORDS },
+    { "escape_host_and_path_string", (PyCFunction)pygvfs_escape_host_and_path_string,
+      METH_VARARGS|METH_KEYWORDS },
+    { "escape_slashes", (PyCFunction)pygvfs_escape_slashes,
+      METH_VARARGS|METH_KEYWORDS },
+    { "escape_set", (PyCFunction)pygvfs_escape_set,
+      METH_VARARGS|METH_KEYWORDS },
+    { "unescape_string", (PyCFunction)pygvfs_unescape_string,
+      METH_VARARGS|METH_KEYWORDS },
+    { "make_uri_canonical", (PyCFunction)pygvfs_make_uri_canonical,
+      METH_VARARGS|METH_KEYWORDS },
+    { "make_path_name_canonical", (PyCFunction)pygvfs_make_path_name_canonical,
+      METH_VARARGS|METH_KEYWORDS },
+    { "unescape_string_for_display", (PyCFunction)pygvfs_unescape_string_for_display,
+      METH_VARARGS|METH_KEYWORDS },
+    { "get_local_path_from_uri", (PyCFunction)pygvfs_get_local_path_from_uri,
+      METH_VARARGS|METH_KEYWORDS },
+    { "get_uri_from_local_path", (PyCFunction)pygvfs_get_uri_from_local_path,
+      METH_VARARGS|METH_KEYWORDS },
+    { "is_executable_command_string", (PyCFunction)pygvfs_is_executable_command_string,
+      METH_VARARGS|METH_KEYWORDS },
+    { "get_volume_free_space", (PyCFunction)pygvfs_get_volume_free_space,
+      METH_VARARGS|METH_KEYWORDS },
+    { "icon_path_from_filename", (PyCFunction)pygvfs_icon_path_from_filename,
+      METH_VARARGS|METH_KEYWORDS },
+    { "open_fd", (PyCFunction)pygvfs_open_fd,
+      METH_VARARGS|METH_KEYWORDS },
+    { "is_primary_thread", (PyCFunction)pygvfs_is_primary_thread,
+      METH_NOARGS },
+    { "format_uri_for_display", (PyCFunction)pygvfs_format_uri_for_display,
+      METH_VARARGS|METH_KEYWORDS },
+    { "make_uri_from_input", (PyCFunction)pygvfs_make_uri_from_input,
+      METH_VARARGS|METH_KEYWORDS },
+    { "make_uri_from_input_with_dirs", (PyCFunction)pygvfs_make_uri_from_input_with_dirs,
+      METH_VARARGS|METH_KEYWORDS },
+    { "make_uri_canonical_strip_fragment", (PyCFunction)pygvfs_make_uri_canonical_strip_fragment,
+      METH_VARARGS|METH_KEYWORDS },
+    { "uris_match", (PyCFunction)pygvfs_uris_match,
+      METH_VARARGS|METH_KEYWORDS },
+    { "get_uri_scheme", (PyCFunction)pygvfs_get_uri_scheme,
+      METH_VARARGS|METH_KEYWORDS },
+    { "make_uri_from_shell_arg", (PyCFunction)pygvfs_make_uri_from_shell_arg,
+      METH_VARARGS|METH_KEYWORDS },
+    { "url_show", (PyCFunction)pygvfs_url_show,
+      METH_VARARGS|METH_KEYWORDS },
+
     { NULL, NULL, 0 }
 };
 
@@ -1779,6 +2146,8 @@ register_constants(PyObject *module)
     regconst(XFER_PHASE_FILECOMPLETED);
     regconst(XFER_PHASE_CLEANUP);
     regconst(XFER_PHASE_COMPLETED);
+    regconst(DIRECTORY_KIND_DESKTOP);
+    regconst(DIRECTORY_KIND_TRASH);
 #undef regconst
 }
 
